unit DDraw;

// Code tidied by Dominique Louis November 1999
// Changed name from DirectDraw to DDraw to avoid naming conflict with Direct 7
// Headers. 26/06/2000.

interface

uses Forms, Dialogs, SysUtils, DirectDraw, Windows, Trace, Utility, Classes;

const
  BIT_FOR_PIXEL : integer = 8;
  SCREEN_WIDTH : integer = 640;
  SCREEN_HEIGHT : integer = 480;

  MAX_SURFACES = 30000;

          // Ub messi in questo modo (+comodo)
  _N = 0;
  _NE = 1;
  _E = 2;
  _SE = 3;
  _S = 4;
  _SW = 5;
  _W = 6;
  _NW = 7;
  STAYING = 0;
  WALKING = 8;
  ATTACK_HIGH = 16;
  ATTACK_MEDIUM = 24;
  ATTACK_LOW = 32;
  FAILED_HIGH = 40;
  FAILED_MEDIUM = 48;
  FAILED_LOW = 56;
  WAITING = 80;

  //Ub per i successivi bisogna dividere x 2 la direzione
  DYING = 64;
  DIED = 68;
  CLIMBING = 72;
  DESCENDING = 76;
  //Ub icona per l'interfaccia
  ICON = 88; //sempre l'ultima

  //costanti ambientali
  AMB_NE = 0;
  AMB_SE = 1;
  AMB_SW = 2;
  AMB_NW = 3;
  DAMAGING = 4;
  DAMAGED = 8;
  DESTROYING = 12;
  DESTROYED = 16;

type
  TScreenMode = record
    Width : integer;
    Height : integer;
    BitsPerPixels : integer;
  end;

  TSurfaceData = record
    BitmapName : string;
    Surface : IDirectDrawSurface;
  end;

  TSurfacesData = array[ 0..MAX_SURFACES ] of TSurfaceData;
  PSurfacesData = ^TSurfacesData;

  {TDDSurfaces  un array dinamico di strutture TSurfaceData.
  Ogni struttura contiene un riferimento ad una superficie DirectDraw e il
  nome del bitmap da cui  stata creata la relativa superficie.
  Come usare TDDSurfaces:
  - creare l'oggetto passando la dimensione iniziale voluta,
  - chiamare SearchBitmap() per verificare che la superifice che si
   vuole creare non esista gi,
  - se la superifice non esiste chiamare Add() per aggiungerla.}
  TDDSurfaces = class
  private
    FSize : Integer; // dimensioni in n di TSurfaceData
    FpData : PSurfacesData;
    FAllocationUnit : Integer;
    FLastItem : Integer;
  public
    // funzioni generali
    constructor Create( Size : Integer );
    destructor Destroy; override;
    procedure SetSize( Size : Integer ); // dimensioni in n di TSurfaceData
    function GetSize : Integer;
    procedure SetAllocationUnit( AllocationUnit : Integer );
    function GetAllocationUnit : Integer;
    procedure SetValue( Index : Integer; Value : TSurfaceData );
    function GetValue( Index : Integer ) : TSurfaceData;
    function AddValue( Value : TSurfaceData ) : Integer;
    procedure Pack;
    // funzioni specifiche per le superifici DD
    function SearchBitmap( BitmapName : string ) : IDirectDrawSurface;
  end;

  {TDirectDraw contiene:
  - oggetto DirectDraw,
  - doppio o triplo buffer (superficie "anteriore" e "posteriore"),
  - array di superifici DirectDraw,
  - palette}
  TDirectDraw = class
  private
    FHandle : THandle;
    FScreenMode : TScreenMode;
  public
    FDirectDraw : IDirectDraw2;
    FFrontSurface : IDirectDrawSurface;
    FBackSurface : IDirectDrawSurface;
    FPalette : IDirectDrawPalette;
    FSurfaces : TDDSurfaces;
    constructor Create( Handle : THandle; ForceDoubleBuffer : Boolean; ScreenMode : TScreenMode );
    destructor Destroy; override;
    procedure Restore;
    function LoadBitmap( BitmapName : string; UseSysMemory : Boolean ) : IDirectDrawSurface;
    procedure LoadPalette( BitmapName : string );
    property CurrScreenMode : TScreenMode read FScreenMode;
  end;

procedure DDCheck( ReturnValue : HRESULT );

function ScreenMode( W, H, BpP : integer ) : TScreenMode;

implementation

uses DirectX;
/////////////////////////////////////////////////////////////////////////
// TSurfacesArray funzioni generali

constructor TDDSurfaces.Create( Size : Integer );
begin
  if Size > MAX_SURFACES then
    raise Exception.Create( 'Invalid size in dynamic array' );
  GetMem( FpData, Size * SizeOf( TSurfaceData ) ); //Ub GetMem solleva un eccezione se salta
  FillChar( FpData^, Size * SizeOf( TSurfaceData ), 0 ); //riempie di zeri

  //Cor NON usare GetMem perch il puntatore che ritorna non  valido. Non so perch.
  //FpData := AllocMem(Size * SizeOf(TSurfaceData));
  FSize := Size;
  FLastItem := 0;
  FAllocationUnit := 10;
end;

destructor TDDSurfaces.Destroy;
var
  nCount : Integer;
begin
  for nCount := 1 to FLastItem do
    FpData^[ nCount ].Surface._Release;
  FreeMem( FpData, FSize * SizeOf( TSurfaceData ) );
  inherited;
end;

procedure TDDSurfaces.SetSize( Size : Integer ); // dimensioni in n di TSurefaceData
var
  {OldFpData,} pData : PSurfacesData;
begin
  //TraceString('Set size start');
  if ( Size > MAX_SURFACES ) or ( Size < FLastItem + 1 ) then
    raise Exception.Create( 'Invalid size in dynamic array' );
  if Size = FSize then Exit;

  { GetMem(pData, Size * SizeOf(TSurfaceData)); //Ub GetMem solleva un eccezione se salta
 //Cor NON usare GetMem perch il puntatore che ritorna non  valido. Non so perch.
  //	pData := AllocMem(Size * SizeOf(TSurfaceData));
  //	FpData:=pData;
  //	Move (FpData, pData, Min(Size, FSize) * SizeOf(TSurfaceData));
        Move (FpData^, pData^, FSize * SizeOf(TSurfaceData)); //move porta source=dest
        FreeMem (FpData, FSize * SizeOf(TSurfaceData));
        FpData:=pData;}

  pData := FpData;
  FpData := nil;
  GetMem( FpData, Size * SizeOf( TSurfaceData ) ); //Ub GetMem solleva un eccezione se salta
  FillChar( FpData^, Size * SizeOf( TSurfaceData ), 0 ); //riempie di zeri
  Move( pData^, FpData^, FSize * SizeOf( TSurfaceData ) ); //move porta source=dest
  FreeMem( pData, FSize * SizeOf( TSurfaceData ) );

  FSize := Size;
  //TraceString('set size end');
end;

function TDDSurfaces.GetSize : Integer;
begin
  Result := FSize;
end;

procedure TDDSurfaces.SetAllocationUnit( AllocationUnit : Integer );
begin
  if AllocationUnit < 1 then
    raise Exception.Create( 'Invalid allocation unit' );
  FAllocationUnit := AllocationUnit;
end;

function TDDSurfaces.GetAllocationUnit : Integer;
begin
  Result := FAllocationUnit;
end;

procedure TDDSurfaces.SetValue( Index : Integer; Value : TSurfaceData );
begin
  if ( Index < 1 ) or ( Index > FLastItem ) then
    raise Exception.Create( 'Index out of bounds in dynamic array' );
  FpData^[ Index ] := Value;
end;

function TDDSurfaces.GetValue( Index : Integer ) : TSurfaceData;
begin
  if ( Index < 1 ) or ( Index > FlastItem ) then
    raise Exception.Create( 'Index out of bounds in dynamic array' );
  Result := FpData^[ Index ];
end;

{AddValue aggiunge un elemento in coda all'array. Se l'array  pieno
 ne aumenta prima le dimensioni di FAllocationUnit unit}

function TDDSurfaces.AddValue( Value : TSurfaceData ) : Integer;
begin
  if FLastItem + 1 >= FSize then //mancava un = (FSize e' zero based)
    SetSize( FSize + FAllocationUnit );
//	TraceString(inttostr(FLastItem));
  FpData^[ FLastItem + 1 ] := Value;
//	TraceString(inttostr(FLastItem));
  FLastItem := FLastItem + 1;
  Result := FLastItem;
end;

{Pack porta l'array alla minima dimensione sufficiente per contenere gli
 elementi attuali}

procedure TDDSurfaces.Pack;
begin
  SetSize( FLastItem );
end;

// funzioni specifiche per le superifici DD
{SearchBitmap cerca se la superficie DD che si vuole creare esiste gi.
 La ricerca viene fatta con un confronto, non sensibile alle maiuscole, sul
 nome del file bitmap che ha originato la superficie. Se il confronto  positivo
 viene restituito il riferimento a IDirectDrawSurface, altrimenti nil}

function TDDSurfaces.SearchBitmap( BitmapName : string ) : IDirectDrawSurface;
var
  nCount : Integer;
begin
  Result := nil;
  for nCount := 1 to FLastItem do
    if UpperCase( FpData^[ nCount ].BitmapName ) = UpperCase( BitmapName ) then
    begin
      Result := FpData^[ nCount ].Surface;
      Break;
    end;
end;

/////////////////////////////////////////////////////////////////////////
// TDirectDraw public methods

destructor TDirectDraw.Destroy;
begin
  FSurfaces.Free;
  if FFrontSurface <> nil then
  begin
    DDCheck( FFrontSurface._Release );
    FFrontSurface := nil;
  end;
  if FPalette <> nil then
  begin
    DDCheck( FPalette._Release );
    FPalette := nil;
  end;
  if FDirectDraw <> nil then
  begin
    DDCheck( FDirectDraw._Release );
    FDirectDraw := nil;
  end;
end;

constructor TDirectDraw.Create( Handle : THandle; ForceDoubleBuffer : Boolean; ScreenMode : TScreenMode );
var
  Caps : TDDSCAPS;
  SurfaceDesc : TDDSURFACEDESC;
  TotMem, FreeMem : Cardinal;
  DirectDraw : IDirectDraw;
begin
  // Initialize the DirectDraw system
  DDCheck( DirectDrawCreate( nil, DirectDraw, nil ) );
  DDCheck( DirectDraw.QueryInterface( IID_IDirectDraw2, FDirectDraw ) );
  TraceString( 'DirectDraw system initialized' );

  // Set full screen and exclusive, so we can flip video pages
  FHandle := Handle;
  DDCheck( FDirectDraw.SetCooperativeLevel( FHandle,
    DDSCL_FULLSCREEN or DDSCL_EXCLUSIVE or DDSCL_ALLOWREBOOT ) );
  TraceString( 'Full screen and exclusive initialized' );

  // Set display mode
  DDCheck( FDirectDraw.SetDisplayMode( ScreenMode.Width, ScreenMode.Height, ScreenMode.BitsPerPixels, 0, 0 ) );
  FScreenMode := ScreenMode;
  TraceString( 'Display mode initialized' );

  // Get available video memory
  //Ub provato la riga sotto DDCheck (FDirectDraw.GetAvailableVidMem (Caps, TotMem, FreeMem));
  if FDirectDraw.GetAvailableVidMem( Caps, TotMem, FreeMem ) <> DD_OK then
  begin
    TotMem := 0;
    FreeMem := 0;
    TraceString( 'Cannot get Available VidMem' );
    TraceString( 'Try to continue as it was never asked for' );
  end
  else
    TraceString( 'Available VidMem Got' );

  // Create flipping surfaces
  with SurfaceDesc do
  begin
    dwSize := SizeOf( SurfaceDesc );
    dwFlags := DDSD_CAPS or DDSD_BACKBUFFERCOUNT;
    if ( FreeMem < 1000000 ) or ForceDoubleBuffer then
    begin
      dwBackBufferCount := 1; //double buffering
      TraceString( 'Double buffering set' ); //Ub
    end
    else
    begin
      dwBackBufferCount := 2; //triple buffering
      TraceString( 'Triple buffering set' ); //Ub
    end;
    ddsCaps.dwCaps := DDSCAPS_PRIMARYSURFACE or DDSCAPS_FLIP or DDSCAPS_COMPLEX;
  end;
  DDCheck( FDirectDraw.CreateSurface( SurfaceDesc, FFrontSurface, nil ) );
  Caps.dwCaps := DDSCAPS_BACKBUFFER; // or DDCAPS_PALETTE; //esp

  DDCheck( FFrontSurface.GetAttachedSurface( Caps, FBackSurface ) );
  TraceString( 'Flipping surfaces initialized' );

  // inizializzo FSurfaces
  FSurfaces := TDDSurfaces.Create( 100 ); //100 per iniziare
  TraceString( 'Surfaces created' );
end;

procedure TDirectDraw.Restore;
begin
  DDCheck( FDirectDraw.SetCooperativeLevel( FHandle, DDSCL_FULLSCREEN or DDSCL_EXCLUSIVE or
    DDSCL_ALLOWREBOOT ) );
  DDCheck( FDirectDraw.SetDisplayMode( CurrScreenMode.Width, CurrScreenMode.Height, CurrScreenMode.BitsPerPixels, 0, 0 ) );
  DDCheck( FFrontSurface._Restore );
end;

function TDirectDraw.LoadBitmap( BitmapName : string; UseSysMemory : Boolean ) : IDirectDrawSurface;
var
  hBmp : HBITMAP;
  Bmp : TBITMAP;
  hdcBmp : HDC;
  SurfDesc : TDDSURFACEDESC;
  hdcSurf : HDC;
  SurfaceData : TSurfaceData;
  //PixelForm: ddPixelFormat;
begin
  // verifica se il bmp  gi caricato in una superfice DD
  Result := FSurfaces.SearchBitmap( BitmapName );
  if Result <> nil then
    Exit;

  // se non trovi il bmp caricalo come oggetto GDI
  hBmp := LoadImage( 0,
    PChar( BitmapName ),
    IMAGE_BITMAP,
    0,
    0,
    LR_LOADFROMFILE or LR_CREATEDIBSECTION );
  if hBmp = 0 then
    raise Exception.Create( 'Unable to load ' + BitmapName );

  // crea un DirectDrawSurface delle dimensioni giuste
  FillChar( SurfDesc, sizeof( SurfDesc ), 0 );
  GetObject( hBmp, sizeof( Bmp ), @Bmp ); // get size of bitmap
  with SurfDesc do
  begin
    dwSize := sizeof( SurfDesc );
    dwFlags := DDSD_CAPS or DDSD_HEIGHT or DDSD_WIDTH;
    {dwFlags := dwFlags + DDSD_PIXELFORMAT; //X
    ddpfPixelFormat.dwSize := SizeOf(DDPixelFormat); //X
    ddpfPixelFormat.dwFlags := DDPF_PALETTEINDEXED8; //X
    ddsCaps.dwCaps := DDSCAPS_TEXTURE; //X}
    ddsCaps.dwCaps := DDSCAPS_OFFSCREENPLAIN;

    if UseSysMemory then
      ddsCaps.dwCaps := ddsCaps.dwCaps + DDSCAPS_SYSTEMMEMORY;
    dwWidth := Bmp.bmWidth;
    dwHeight := Bmp.bmHeight;
  end;
  DDCheck( FDirectDraw.CreateSurface( SurfDesc, SurfaceData.Surface, nil ) );

  // ottieni i DC per il bitmap GDI e per DirectDrawSurface
  DDCheck( SurfaceData.Surface.GetDC( hdcSurf ) );
  hdcBmp := CreateCompatibleDC( hdcSurf );
  if hdcBmp = 0 then
    raise Exception.Create( 'Unable to CreateCompatibleDC' );
  SelectObject( hdcBmp, hBmp );

  // copia il bitmap GDI su DirectDrawSurface
  BitBlt( hdcSurf, 0, 0, Bmp.bmWidth, Bmp.bmHeight, hdcBmp, 0, 0, SRCCOPY );

  // rilascia le risorse
  DDCheck( SurfaceData.Surface.ReleaseDC( hdcSurf ) );
  DeleteDC( hdcBmp );
  DeleteObject( hBmp );

  //adesso salva tutto in FSurfaces
  SurfaceData.BitmapName := BitmapName;
  FSurfaces.AddValue( SurfaceData );
  Result := SurfaceData.Surface;
end;

procedure TDirectDraw.LoadPalette( BitmapName : string );
var
  BmpFile : TFileStream;
  BmpFileHeader : TBITMAPFILEHEADER;
  BmpInfoHeader : TBITMAPINFOHEADER;
  BmpPalette : array[ 0..255 ] of TPALETTEENTRY;
  nCount : Integer;
  AColor : BYTE;
begin
  BmpFile := TFileStream.Create( BitmapName, fmOpenRead or fmShareDenyWrite );
  with BmpFile do
  begin
    Read( BmpFileHeader, SizeOf( BmpFileHeader ) );
    Read( BmpInfoHeader, SizeOf( BmpInfoHeader ) );
    Read( BmpPalette, SizeOf( BmpPalette ) );
    Free;
  end;
// a DIB color table has its colors stored BGR not RGB so flip them.
  for nCount := 0 to 255 do
  begin
    AColor := BmpPalette[ nCount ].peRed;
    BmpPalette[ nCount ].peRed := BmpPalette[ nCount ].peBlue;
    BmpPalette[ nCount ].peBlue := AColor;
  end;

  DDCheck( FDirectDraw.CreatePalette( DDPCAPS_8BIT, @BmpPalette, FPalette, nil ) );
  if CurrScreenMode.BitsPerPixels = 8 then
    DDCheck( FFrontSurface.SetPalette( FPalette ) ); //carico la palette sulla FrontSurface solo se siamo a 8 bit
end;

/////////////////////////////////////////////////////////////////////////
// funzioni di utilit

procedure DDCheck( ReturnValue : HRESULT );
var
  ErrorMsg : string;
begin
  if ReturnValue = DD_OK then exit;
  case ReturnValue of
    {DD_OK: S:= 'No error';}
    DDERR_ALREADYINITIALIZED : ErrorMsg := 'This object is already initialized';
    DDERR_BLTFASTCANTCLIP : ErrorMsg := ' if a clipper object is attached to the source surface passed into a BltFast call';
    DDERR_CANNOTATTACHSURFACE : ErrorMsg := 'This surface can not be attached to the requested surface';
    DDERR_CANNOTDETACHSURFACE : ErrorMsg := 'This surface can not be detached from the requested surface';
    DDERR_CANTCREATEDC : ErrorMsg := 'Windows can not create any more DCs';
    DDERR_CANTDUPLICATE : ErrorMsg := 'Cannot duplicate primary & 3D surfaces, or surfaces that are implicitly created';
    DDERR_CLIPPERISUSINGHWND : ErrorMsg := 'An attempt was made to set a cliplist for a clipper object that is already monitoring an hwnd';
    DDERR_COLORKEYNOTSET : ErrorMsg := 'No src color key specified for this operation';
    DDERR_CURRENTLYNOTAVAIL : ErrorMsg := 'Support is currently not available';
    DDERR_DIRECTDRAWALREADYCREATED : ErrorMsg := 'A DirectDraw object representing this driver has already been created for this process';
    DDERR_EXCEPTION : ErrorMsg := 'An exception was encountered while performing the requested operation';
    DDERR_EXCLUSIVEMODEALREADYSET : ErrorMsg := 'An attempt was made to set the cooperative level when it was already set to exclusive';
    DDERR_GENERIC : ErrorMsg := 'Generic failure';
    DDERR_HEIGHTALIGN : ErrorMsg := 'Height of rectangle provided is not a multiple of reqd alignment';
    DDERR_HWNDALREADYSET : ErrorMsg := 'The CooperativeLevel HWND has already been set. It can not be reset while the process has surfaces or palettes created';
    DDERR_HWNDSUBCLASSED : ErrorMsg := 'HWND used by DirectDraw CooperativeLevel has been subclassed, this prevents DirectDraw from restoring state';
    DDERR_IMPLICITLYCREATED : ErrorMsg := 'This surface can not be restored because it is an implicitly created surface';
    DDERR_INCOMPATIBLEPRIMARY : ErrorMsg := 'Unable to match primary surface creation request with existing primary surface';
    DDERR_INVALIDCAPS : ErrorMsg := 'One or more of the caps bits passed to the callback are incorrect';
    DDERR_INVALIDCLIPLIST : ErrorMsg := 'DirectDraw does not support the provided cliplist';
    DDERR_INVALIDDIRECTDRAWGUID : ErrorMsg := 'The GUID passed to DirectDrawCreate is not a valid DirectDraw driver identifier';
    DDERR_INVALIDMODE : ErrorMsg := 'DirectDraw does not support the requested mode';
    DDERR_INVALIDOBJECT : ErrorMsg := 'DirectDraw received a pointer that was an invalid DIRECTDRAW object';
    DDERR_INVALIDPARAMS : ErrorMsg := 'One or more of the parameters passed to the function are incorrect';
    DDERR_INVALIDPIXELFORMAT : ErrorMsg := 'The pixel format was invalid as specified';
    DDERR_INVALIDPOSITION : ErrorMsg := 'Returned when the position of the overlay on the destination is no longer legal for that destination';
    DDERR_INVALIDRECT : ErrorMsg := 'Rectangle provided was invalid';
    DDERR_LOCKEDSURFACES : ErrorMsg := 'Operation could not be carried out because one or more surfaces are locked';
    DDERR_NO3D : ErrorMsg := 'There is no 3D present';
    DDERR_NOALPHAHW : ErrorMsg := 'Operation could not be carried out because there is no alpha accleration hardware present or available';
    DDERR_NOBLTHW : ErrorMsg := 'No blitter hardware present';
    DDERR_NOCLIPLIST : ErrorMsg := 'No cliplist available';
    DDERR_NOCLIPPERATTACHED : ErrorMsg := 'No clipper object attached to surface object';
    DDERR_NOCOLORCONVHW : ErrorMsg := 'Operation could not be carried out because there is no color conversion hardware present or available';
    DDERR_NOCOLORKEY : ErrorMsg := 'Surface does not currently have a color key';
    DDERR_NOCOLORKEYHW : ErrorMsg := 'Operation could not be carried out because there is no hardware support of the destination color key';
    DDERR_NOCOOPERATIVELEVELSET : ErrorMsg := 'Create function called without DirectDraw object method SetCooperativeLevel being called';
    DDERR_NODC : ErrorMsg := 'No DC was ever created for this surface';
    DDERR_NODDROPSHW : ErrorMsg := 'No DirectDraw ROP hardware';
    DDERR_NODIRECTDRAWHW : ErrorMsg := 'A hardware-only DirectDraw object creation was attempted but the driver did not support any hardware';
    DDERR_NOEMULATION : ErrorMsg := 'Software emulation not available';
    DDERR_NOEXCLUSIVEMODE : ErrorMsg := 'Operation requires the application to have exclusive mode but the application does not have exclusive mode';
    DDERR_NOFLIPHW : ErrorMsg := 'Flipping visible surfaces is not supported';
    DDERR_NOGDI : ErrorMsg := 'There is no GDI present';
    DDERR_NOHWND : ErrorMsg := 'Clipper notification requires an HWND or no HWND has previously been set as the CooperativeLevel HWND';
    DDERR_NOMIRRORHW : ErrorMsg := 'Operation could not be carried out because there is no hardware present or available';
    DDERR_NOOVERLAYDEST : ErrorMsg := 'Returned when GetOverlayPosition is called on an overlay that UpdateOverlay has never been called on to establish a destination';
    DDERR_NOOVERLAYHW : ErrorMsg := 'Operation could not be carried out because there is no overlay hardware present or available';
    DDERR_NOPALETTEATTACHED : ErrorMsg := 'No palette object attached to this surface';
    DDERR_NOPALETTEHW : ErrorMsg := 'No hardware support for 16 or 256 color palettes';
    DDERR_NORASTEROPHW : ErrorMsg := 'Operation could not be carried out because there is no appropriate raster op hardware present or available';
    DDERR_NOROTATIONHW : ErrorMsg := 'Operation could not be carried out because there is no rotation hardware present or available';
    DDERR_NOSTRETCHHW : ErrorMsg := 'Operation could not be carried out because there is no hardware support for stretching';
    DDERR_NOT4BITCOLOR : ErrorMsg := 'DirectDrawSurface is not in 4 bit color palette and the requested operation requires 4 bit color palette';
    DDERR_NOT4BITCOLORINDEX : ErrorMsg := 'DirectDrawSurface is not in 4 bit color index palette and the requested operation requires 4 bit color index palette';
    DDERR_NOT8BITCOLOR : ErrorMsg := 'DirectDrawSurface is not in 8 bit color mode and the requested operation requires 8 bit color';
    DDERR_NOTAOVERLAYSURFACE : ErrorMsg := 'Returned when an overlay member is called for a non-overlay surface';
    DDERR_NOTEXTUREHW : ErrorMsg := 'Operation could not be carried out because there is no texture mapping hardware present or available';
    DDERR_NOTFLIPPABLE : ErrorMsg := 'An attempt has been made to flip a surface that is not flippable';
    DDERR_NOTFOUND : ErrorMsg := 'Requested item was not found';
    DDERR_NOTLOCKED : ErrorMsg := 'Surface was not locked.  An attempt to unlock a surface that was not locked at all, or by this process, has been attempted';
    DDERR_NOTPALETTIZED : ErrorMsg := 'The surface being used is not a palette-based surface';
    DDERR_NOVSYNCHW : ErrorMsg := 'Operation could not be carried out because there is no hardware support for vertical blank synchronized operations';
    DDERR_NOZBUFFERHW : ErrorMsg := 'Operation could not be carried out because there is no hardware support for zbuffer blitting';
    DDERR_NOZOVERLAYHW : ErrorMsg := 'Overlay surfaces could not be z layered based on their BltOrder because the hardware does not support z layering of overlays';
    DDERR_OUTOFCAPS : ErrorMsg := 'The hardware needed for the requested operation has already been allocated';
    DDERR_OUTOFMEMORY : ErrorMsg := 'DirectDraw does not have enough memory to perform the operation';
    DDERR_OUTOFVIDEOMEMORY : ErrorMsg := 'DirectDraw does not have enough memory to perform the operation';
    DDERR_OVERLAYCANTCLIP : ErrorMsg := 'The hardware does not support clipped overlays';
    DDERR_OVERLAYCOLORKEYONLYONEACTIVE : ErrorMsg := 'Can only have ony color key active at one time for overlays';
    DDERR_OVERLAYNOTVISIBLE : ErrorMsg := 'Returned when GetOverlayPosition is called on a hidden overlay';
    DDERR_PALETTEBUSY : ErrorMsg := 'Access to this palette is being refused because the palette is already locked by another thread';
    DDERR_PRIMARYSURFACEALREADYEXISTS : ErrorMsg := 'This process already has created a primary surface';
    DDERR_REGIONTOOSMALL : ErrorMsg := 'Region passed to Clipper::GetClipList is too small';
    DDERR_SURFACEALREADYATTACHED : ErrorMsg := 'This surface is already attached to the surface it is being attached to';
    DDERR_SURFACEALREADYDEPENDENT : ErrorMsg := 'This surface is already a dependency of the surface it is being made a dependency of';
    DDERR_SURFACEBUSY : ErrorMsg := 'Access to this surface is being refused because the surface is already locked by another thread';
    DDERR_SURFACEISOBSCURED : ErrorMsg := 'Access to surface refused because the surface is obscured';
    DDERR_SURFACELOST : ErrorMsg := 'Access to this surface is being refused because the surface memory is gone. The DirectDrawSurface object representing this surface should have Restore called on it';
    DDERR_SURFACENOTATTACHED : ErrorMsg := 'The requested surface is not attached';
    DDERR_TOOBIGHEIGHT : ErrorMsg := 'Height requested by DirectDraw is too large';
    DDERR_TOOBIGSIZE : ErrorMsg := 'Size requested by DirectDraw is too large, but the individual height and width are OK';
    DDERR_TOOBIGWIDTH : ErrorMsg := 'Width requested by DirectDraw is too large';
    DDERR_UNSUPPORTED : ErrorMsg := 'Action not supported';
    DDERR_UNSUPPORTEDFORMAT : ErrorMsg := 'FOURCC format requested is unsupported by DirectDraw';
    DDERR_UNSUPPORTEDMASK : ErrorMsg := 'Bitmask in the pixel format requested is unsupported by DirectDraw';
    DDERR_VERTICALBLANKINPROGRESS : ErrorMsg := 'Vertical blank is in progress';
    DDERR_WASSTILLDRAWING : ErrorMsg := 'The previous Blt which is transfering information to or from this Surface is incomplete';
    DDERR_WRONGMODE : ErrorMsg := 'This surface can not be restored because it was created in a different mode';
    DDERR_XALIGN : ErrorMsg := 'Rectangle provided was not horizontally aligned on required boundary';
  end;
  ErrorMsg := 'DirectDraw error hex ' + IntToHex( ReturnValue, 8 ) + ': ' + ErrorMsg;
  TraceString( ErrorMsg ); //ho modificato TraceString e affini
  raise Exception.Create( ErrorMsg ); //bisogna settare BreakOnException a false
end;

function ScreenMode( W, H, BpP : integer ) : TScreenMode;
begin
  result.Width := W;
  result.Height := H;
  result.BitsPerPixels := BpP;
end;

end.
